local object = require("object")
local canBuildFunctions = require("canBuild")
-- Settings for Overlords Update

-- canBuildParameters
--      Three tables, one for unitTypes, one for Improvements, one for Wonders
--      absent entry means use the defaultCanBuild function
-- canBuildObjectType[item.id]= {
--      .forbiddenTribes = {[tribeID]=bool}
--          if canBuildObjectType[item.id].forbiddenTribes[tribeID] is true, then the tribe with
--          tribeID can't build item, false or nil/absent means it can
--
--      .forbiddenMaps = {[0] = bool,[1]=bool,[2]=bool,[3]=bool}
--          if canBuildObjectType[item.id].forbiddenMaps[mapCityIsOn] = true, then city can't build the item
--              false or nil means it can
--          absent means all maps are allowed
--      .location = {xCoord,yCoord} or {xCoord,yCoord,zCoord} or tileObject or cityObject or integer or function(tileObject)-->boolean or table of these kinds of objects
--          {xCoord,yCoord} if the city is located at (xCoord,yCoord) on any map, it can build the object
--          {xCoord,yCoord,zCoord} means the city must be located at those coordinates to build the object
--          tileObject means the city must be located at that tile
--          cityObject means the city must be that city
--          function means object can be built if function(city.location) returns true 
--          (and all other conditions are met), and can't be built otherwise
--          table of these things means that each entry in the table is checked, and if any one of them means the object can be built, then it can be built
--          absent means the object can be built at any location
--          (Note: Can't use integers to match city id, since code can't distinguish between several cities and a coordinate triple)
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .forbiddenLocation= {xCoord,yCoord} or {xCoord,yCoord,zCoord} or tileObject or cityObject or function(tileObject)-->boolean or table of these kinds of objects
--              see location details, except that a match in forbidden location prevents the item from being buitl
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .allImprovements = improvementObject or wonderObject or table of improvementObjects and wonderObjects
--          The city must have all improvements/wonders in the table to build the object
--          absent means no improvements needed (in this section)
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .someImprovements = improvementObject or wonderObject or table of improvementObjects and wonderObjects
--          The city must have some number of objects in the table to build the item in question.  The exact
--          number is governed by the next key
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .numberOfImprovements = integer
--          tells how many of the 'someImprovements' are needed to build the item
--          absent means ignore .someImprovements
--      .forbiddenImprovements = improvementObject or wonderObject or table of improvementObjects and wonderObjects
--          The city must have none of the improvements/wonders in the table to build the object
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .allTechs = technologyObject or table of technologyObjects
--          the civ must have all the technologies in the table to build the object
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .someTechs = technologies or table of technologyObjects
--          the civ must have some of the technologies in the table to build the object
--          the number is governed by the next key
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .numberOfTechs = integer
--          tells how many of the 'someTechs' are needed for the object to be built
--      .forbiddenTechs = technologyObject or table of technologyObjects
--          the civ must not have any of the technologies in the table to build the object
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .allFlagsMatch = {[flagKey]=boolean}
--          the city can only build the item if all the flags for the flagKeys in the table have the  
--          corresponding value
--          absent flag key (including [flagKey]=nil) doesn't affect production
--          absent means no restriction
--      .someFlagsMatch = {[flagKey]=boolean}
--          the city can only build the item if all the flags for the flagKeys in the table have the  
--          corresponding value
--          absent flag key (including [flagKey]=nil) doesn't affect production
--          absent means no restriction
--      .numberOfFlags = integer
--          tells how many of the 'someFlags' are needed for the object to be built
--      .minimumPopulation = integer
--          the city must have at least this many citizens to build the item
--          absent means 0
--      .maximumPopulation = integer 
--          the city can have at most this many citizens to build the item
--          absent means no maximum population
--      .earliestTurn = integer
--          item can't be built before this turn
--          absent means no restriction
--      .latestTurn = integer
--          item can't be built after this turn
--          absent means no restriction
--      .allWonders = wonderObject or table of wonderObject
--          the city's tribe must have all these wonders in order to build the item
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .someWonders = wonderObject or table of wonderObject
--          the city's tribe must have some of the wonders in order to build the item
--          the number is governed by the next key
--          absent means no restriction
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .numberOfWonders = integer
--          tells how many of 'someWonders' the tribe must have to build the item
--          absent means no restriction
--
--
--      .overrideDefaultBuildFunction = boolean or nil
--          if true, the in game function for determining if the item can be built is ignored
--          for improvements and wonders, a check will be made if the item has already been built
--          false, nil, absent means the game's regular conditions must also be met
--      .forbiddenAlternateProduction = unitTypeObject or imprvementObject or wonderObject or table of these objects
--          the city can only build the item if city:canBuild(item) returns false for all items in the list
--          Be careful that the 'can build chain' does not form a loop
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .requireSomeAsAlternateProduction = unitTypeObject or improvementObject or wonderObject or table of these objects
--          the city can only build the item if the city can build some of the items in the list
--          Be careful that the 'can build chain' does not form a loop
--          A single entry not in a table will be 'wrapped' with a table in post processing
--      .numberOfAlternateProduction = integer
--          tells how many of 'someAlternateProduction' is required
--          absent means no restriction
--      .conditionFunction = function(defaultBuildFunction,city,item) --> bool
--          if function returns true, item can be built if other conditions are met, if false, item can't be built
--          absent means no extra condition
--      .returnFalse = bool or nil
--          if true, item can't be built
--          if false or nil, refer to other conditions
--          (happens before overrideFunction and alternateParameters)
--      .overrideFunction = function(defaultBuildFunction,city,item) --> boolean
--          if function returns true, the city is automatically allowed to build the item, regardless of any
--          conditions that isn't met
--          if function returns false, the other conditions are checked
--      .alternateParameters = table of itemParameters
--          itemParameters is this table of restrictions on whether a given item can be produced
--          if the item in question satisfies any of the itemParameters in the table, it can be produced,
--          regardless of whether the 'top' itemParameters are satisfied
--          use this (or overrideFunction) if you want to have more than one valid way to produce the item
--          the 'table' format is important.  Unlike other parameters, you must enclose the value of 
--          alternateParameters in a table, even if there is only one itemParameters as the value
--      .computerOnly = bool or nil
--          if true, item can only be built by computer controlled players
--          if false or nil, either human or AI players can build
--          (in conjunction with alternateParameters, this can be used to have different conditions for the
--          ai and human players)
--      .humanOnly = bool or nil
--          if true, item can only be built by human controlled players
--          if false or nil, either human or AI players can build
--          (in conjunction with alternateParameters, this can be used to have different conditions for the
--          ai and human players)
--
--      .onlyBuildCoastal = bool or nil
--          if true, the item can only be built if the city has the 'coastal' flag,
--          that is, in the default game it could build harbors and offshore platforms
--      .onlyBuildShips = bool or nil
--          if true, the item can only be built if the city has the 'ship building' flag
--          that is, in the default game it could build sea units
--      .onlyBuildHydroPlant = bool or nil
--          if true, the item can only be built if the city has the 'can build hydro plant' flag
--          that is, the city could build hydro plants in the default game
--
--

--Engineer Special Capital Condition
local engineerAltCondition = {allImprovements=object.iCapital, overrideDefaultBuildFunction=true}

--Soviet Red Army Recruitment Locations
local SovietHomeCities = {{21,41,0},{22,20,0},{22,34,0},{22,46,0},{24,30,0},{25,21,0},{25,41,0},{25,47,0},{26,6,0},{26,32,0},{27,13,0},{27,17,0},
{28,46,0},{28,50,0},{30,30,0},{31,21,0},{31,25,0},{31,45,0},{32,36,0},{33,13,0},{33,49,0},{34,42,0},{36,28,0},{36,56,0},{37,51,0},{38,20,0},{39,57,0},
{40,28,0},{40,34,0},{42,52,0},{43,37,0},{43,43,0},{44,58,0},{45,13,0},{46,24,0},{46,36,0},{48,32,0},{49,37,0},{49,53,0},{50,8,0},{51,43,0},{52,24,0},
{52,30,0},{54,62,0},{59,37,0},{59,53,0},{56,58,0},{61,27,0},{62,44,0},{62,50,0},{69,29,0},{73,33,0},{78,28,0},{82,30,0},{82,38,0},{84,44,0},{87,39,0},
{92,38,0},{100,40,0},{103,17,0},{107,51,0},{108,42,0},{117,23,0},{128,34,0}}

--Soviet Airborne Recruitment Locations
local SovAirborneCities = {{20,28,0},{21,23,0},{24,30,0},{25,21,0},{30,30,0},{82,38,0},{107,51,0}}

--Soviet Guards Recruitment Locations
local GuardsHomeCities = {{25,21,0},{34,42,0},{30,30,0},{107,51,0}}


--Japanese Army Recruitment Locations
local JapaneseHomeCities = {{104,82,0},{115,53,0},{115,57,0},{116,62,0},{116,66,0},{113,69,0},{110,72,0},{108,70,0},{107,75,0}}

--Japanese Army Recruitment Locations
local JapaneseArmyCities = {{104,82,0},{115,53,0},{115,57,0},{116,62,0},{116,66,0},{113,69,0},{110,72,0},{108,70,0},{107,75,0},{63,59,0},{67,63,0},
{71,57,0},{72,52,0},{75,73,0},{76,58,0},{82,62,0},{82,74,0},{82,84,0},{84,44,0},{84,66,0},{85,77,0},{86,58,0},{86,72,0},{86,84,0},{87,69,0},{88,88,0},
{89,59,0},{90,76,0},{90,82,0},{90,90,0},{91,67,0},{91,73,0},{91,95,0},{92,88,0},{93,59,0},{94,44,0},{94,64,0},{94,70,0},{94,90,0},{96,68,0},{96,82,0},
{96,86,0},{97,75,0},{97,89,0},{98,72,0},{100,46,0},{100,56,0},{100,84,0},{101,51,0},{103,47,0}}

--Japanese Paratroops Recruitment Locations
local JapaneseParaCities = {{104,82,0},{115,53,0},{115,57,0},{116,62,0},{116,66,0},{113,69,0},{110,72,0},{108,70,0},{107,75,0}}

--Japanese Marines Recruitment Locations
local JapaneseMarinesCities = {{91,95,0},{100,84,0},{104,82,0},{114,86,0},{115,53,0},{115,57,0},{116,62,0},{116,66,0},{113,69,0},{110,72,0},{108,70,0},{107,75,0}}


--German Army Recruitment Locations
local GermanHomeCities = {
{4,36,0},{5,17,0},{5,21,0},{7,27,0},{7,35,0},{7,39,0},{9,13,0},{9,19,0},{9,33,0},{10,30,0},
{10,40,0},{10,44,0},{11,37,0},{13,33,0},{13,41,0},{13,45,0},{14,6,0},{16,32,0}}

--Waffen SS Recruitment Locations
local WaffenSSCities = {{11,37,0}}

--Fallschirmjager Recruitment Locations
local FallschirmHomeCities = {{11,37,0},{7,39,0}}

--Italian Recruitment Locations
local ItalianHomeCities = {{7,51,0},{9,49,0},{9,53,0},{10,74,0},{11,55,0},{11,63,0},{14,60,0}}

--Finn Recruitment Locations
local FinnHomeCities = {{18,16,0},{19,11,0},{19,21,0},{22,20,0}}

--Romanian Recruitment Locations
local RomanianHomeCities = {{22,50,0}}

--French Recruitment Locations
local VichyHomeCities = {{0,50,0},{0,68,0},{2,38,0},{2,42,0},{3,47,0},{3,51,0},{4,66,0},{6,44,0},
{8,68,0},{275,95,0},{276,100,0},{277,85,0},{284,68,0},{281,71,0},{281,77,0},{284,44,0},{286,46,0}}

--Spanish Recruitment Locations
local SpanishHomeCities = {{1,55,0},{283,61,0},{284,58,0},{285,53,0},{287,59,0}}

--Dutch Recruitment Locations
local DutchHomeCities = {{78,124,0},{80,132,0},{84,136,0},{87,141,0},{91,143,0},{96,144,0},{97,139,0}}

--Garrison Recruitment Locations
local GarrisonHomeCities = {{5,121,0},{7,129,0},{9,137,0},{10,24,0},{10,162,0},{11,45,0},{12,68,0},{12,172,0},{14,22,0},{14,14,0},
{14,180,0},{14,186,0},{15,49,0},{15,191,0},{16,10,0},{16,36,0},{16,170,0},{17,41,0},{17,45,0},{17,57,0},{18,50,0},{19,55,0},{20,62,0},
{20,162,0},{21,57,0},{21,67,0},{21,185,0},{22,178,0},{22,190,0},{23,169,0},{25,161,0},{25,185,0},{26,178,0},{27,67,0},{28,156,0},{28,164,0},
{28,172,0},{28,164,0},{30,160,0},{33,157,0},{36,176,0},{37,167,0},{39,153,0},{133,169,0},{199,89,0},{202,82,0},{204,90,0},{207,83,0},{208,96,0},
{208,102,0},{209,89,0},{213,97,0},{216,92,0},{217,99,0},{217,107,0},{219,91,0},{220,102,0},{220,112,0},{222,88,0},{226,128,0},{226,148,0},
{227,91,0},{228,112,0},{230,118,0},{231,95,0},{234,112,0},{237,201,0},{237,123,0},{238,94,0},{240,134,0},{240,194,0},{243,221,0},{244,194,0},
{245,119,0},{246,188,0},{247,123,0},{247,127,0},{248,16,0},{248,174,0},{249,133,0},{249,165,0},{252,172,0},{257,137,0},{259,149,0},{261,143,0},
{271,77,0},{274,104,0},{275,95,0},{276,100,0},{277,85,0},{277,111,0},{277,143,0},{279,115,0},{280,56,0},{280,60,0},{282,64,0},{284,160,0}}

--Muslim Recruitment Locations
local MuslimHomeCities = {{0,68,0},{0,120,0},{2,104,0},{4,66,0},{6,98,0},{6,114,0},{7,107,0},{8,68,0},{8,124,0},{9,103,0},{10,74,0},
{12,108,0},{16,122,0},{19,107,0},{15,77,0},{16,122,0},{18,72,0},{23,73,0},{23,77,0},{23,83,0},{23,113,0},{24,56,0},{24,88,0},{25,95,0},
{25,101,0},{25,107,0},{26,74,0},{26,120,0},{27,59,0},{27,63,0},{27,127,0},{29,111,0},{30,56,0},{30,72,0},{30,76,0},{30,98,0},{31,63,0},
{31,69,0},{32,104,0},{32,134,0},{32,142,0},{33,87,0},{33,93,0},{34,64,0},{34,138,0},{34,110,0},{35,69,0},{36,60,0},{36,92,0},{36,124,0},
{37,85,0},{37,105,0},{38,74,0},{41,67,0},{41,71,0},{41,103,0},{43,75,0},{43,85,0},{45,97,0},{45,109,0},{47,89,0},{48,66,0},{52,74,0},
{54,62,0},{54,70,0},{54,76,0},{54,82,0},{56,58,0},{58,68,0},{59,53,0},{59,73,0},{59,87,0},{62,44,0},{62,50,0},{67,47,0},{73,43,0},
{275,95,0},{276,100,0},{281,71,0},{281,77,0},{283,103,0},{284,68,0},{284,120,0},{286,98,0},{287,113,0}}

--Indian Cities
--{54,76,0},{54,82,0},{55,89,0},{58,68,0},{59,73,0},{59,87,0},{58,94,0},{59,101,0},{60,108,0},{61,79,0},{62,98,0},
--{62,112,0},{63,75,0},{63,89,0},{64,108,0},{65,83,0},{67,99,0},{69,95,0},{70,88,0},{71,83,0},{73,85,0},{74,88,0},

--Muslim Indian Cities
--{54,76,0},{54,82,0},{58,68,0},{59,73,0},

--Burma Cities
--{77,89,0},{77,95,0},{78,82,0},{78,86,0},{78,100,0},{81,111,0},

--Malaysia Cities
--{81,117,0},{83,121,0},{84,126,0},{87,129,0},{90,124,0},{90,134,0},{92,120,0},

--Indochina Cities
--{82,98,0},{82,106,0},{85,93,0},{85,111,0},{87,101,0},{87,107,0},

--Colonial Rifles Recruitment Locations
local ColonialHomeCities = {{46,80,0},{54,76,0},{54,82,0},{55,89,0},{58,94,0},{59,87,0},{59,101,0},{60,108,0},{61,79,0},{62,98,0},
{62,112,0},{63,75,0},{63,89,0},{64,108,0},{65,83,0},{65,117,0},{67,99,0},{69,95,0},{70,78,0},{70,88,0},{71,83,0},{73,85,0},{74,78,0},
{74,88,0},{77,89,0},{77,95,0},{78,82,0},{78,86,0},{78,100,0},{78,124,0},{80,132,0},{81,111,0},{81,117,0},{82,98,0},
{82,106,0},{83,121,0},{84,126,0},{84,136,0},{85,93,0},{85,111,0},{87,101,0},{87,107,0},{87,129,0},{87,141,0},{90,124,0},{90,134,0},
{91,95,0},{91,143,0},{92,120,0},{94,90,0},{96,144,0},{97,139,0},{100,128,0},{100,148,0},{101,117,0},{102,104,0},
{102,112,0},{106,132,0},{113,135,0},{115,143,0},{117,99,0},{121,137,0},{136,112,0},{141,163,0},{184,176,0}}

--American Recruitment Locations
local USAHomeCities = {{98,100,0},{145,83,0},{160,92,0},{175,21,0},{183,27,0},{188,56,0},{189,47,0},{190,62,0},{192,66,0},{192,44,0},
{194,70,0},{195,53,0},{195,59,0},{196,74,0},{196,46,0},{200,44,0},{200,58,0},{200,72,0},{203,47,0},{204,54,0},{205,61,0},{205,69,0},
{207,49,0},{209,75,0},{210,68,0},{211,55,0},{211,61,0},{211,77,0},{213,47,0},{215,61,0},{216,50,0},{216,54,0},{216,68,0},{216,72,0},
{216,76,0},{219,61,0},{220,52,0},{220,70,0},{220,74,0},{221,79,0},{222,56,0},{222,64,0},{223,81,0},{224,58,0},{224,70,0},{225,53,0},
{225,67,0},{226,62,0},{226,66,0},{228,58,0},{229,55,0},{230,52,0}}

--US Airborne Recruitment Locations
local USAirborneCities = {{200,72,0},{220,70,0},{216,54,0},{228,58,0}}

--US Marines Recruitment Locations
local USMarineCities = {{190,62,0},{194,70,0},{196,74,0},{224,70,0},{226,62,0},{229,55,0}}

--US Navy Air Force Recruitment Locations
local USNAFBases = {{98,100,0},{117,99,0},{98,100,0},{145,83,0},{160,92,0}}


--British Home Recruitment Locations
local BritishHomeCities = {{281,29,0},{283,25,0},{283,39,0},{284,34,0},{285,29,0},{286,38,0},{287,35,0}}

--British Paratrooper Recruitment Locations
local BritParaCities = {{284,34,0},{285,29,0}}

--British Commando Recruitment Locations
local BritCommandoCities = {{283,25,0},{287,35,0}}

--Gurkha Recruitment Locations
local GurkhaHomeCities = {{54,82,0},{58,94,0},{59,73,0},{61,79,0},{62,98,0},{62,112,0},{64,108,0},{70,88,0}}

--Anzac Recruitment Locations
local AnzacHomeCities = {{91,175,0},{93,167,0},{93,187,0},{98,162,0},{106,152,0},{107,193,0},{111,197,0},{115,199,0},
{116,158,0},{117,163,0},{117,195,0},{118,206,0},{119,147,0},{119,169,0},{121,177,0},{121,193,0},{123,183,0},{134,182,0},
{135,213,0},{137,209,0},{140,196,0},{140,204,0}}

--Canadian Recruitment Locations
local CanadianHomeCities = {{188,40,0},{192,40,0},{200,32,0},{201,21,0},{201,39,0},{207,41,0},{217,39,0},{223,49,0},
{228,46,0},{231,43,0},{232,48,0},{236,48,0},{243,47,0}}


--Chinese Recruitment Locations
local ChineseHomeCities = {
{63,59,0},{67,63,0},
{71,57,0},{72,52,0},{75,73,0},{76,58,0},
{82,62,0},{82,74,0},{82,84,0},{84,44,0},{84,66,0},{85,77,0},{86,58,0},{86,72,0},{86,84,0},{87,69,0},{88,88,0},{89,59,0},
{90,76,0},{90,82,0},{90,90,0},{91,67,0},{91,73,0},{91,95,0},{92,88,0},{93,59,0},{94,44,0},{94,64,0},{94,70,0},{94,90,0},
{96,68,0},{96,82,0},{96,86,0},{97,75,0},{97,79,0},{98,72,0},
{100,46,0},{100,56,0},{100,84,0},{101,51,0},{103,47,0}}

--Chinese AVG Fighter Training Locations in Kumming and Canton
local ChineseAVGBases = {{82,84,0},{92,88,0}}


local unitTypeBuild = {}


--NAVAL UNITS
unitTypeBuild[object.uFreighter.id] = {allImprovements={object.iShipyard}}
unitTypeBuild[object.uLiner.id] = {allImprovements={object.iShipyard,object.iMilitaryFactory}}
unitTypeBuild[object.uLibertyShip.id] = {allImprovements={object.iShipyard,object.iMilitaryFactory}}
unitTypeBuild[object.uDestroyer.id] = {allImprovements={object.iShipyard,object.iNavalBase}}
unitTypeBuild[object.uCruiser.id] = {allImprovements={object.iShipyard,object.iNavalBase}}
unitTypeBuild[object.uAACruiser.id] = {allImprovements={object.iShipyard,object.iNavalBase}}
unitTypeBuild[object.uAACruiserII.id] = {allImprovements={object.iShipyard,object.iNavalBase,object.iMilitaryFactory}}
unitTypeBuild[object.uBattleship.id] = {allImprovements={object.iShipyard,object.iNavalBase,object.iMilitaryFactory}}
unitTypeBuild[object.uSuperBattleship.id] = {allImprovements={object.iShipyard,object.iNavalBase,object.iMilitaryFactory}}
unitTypeBuild[object.uAircraftCarrier.id] = {allImprovements={object.iShipyard,object.iNavalBase,object.iMilitaryFactory}}
unitTypeBuild[object.uEscortCarrier.id] = {allImprovements={object.iShipyard,object.iNavalBase,object.iMilitaryFactory}}
unitTypeBuild[object.uSubmarine.id] = {allImprovements={object.iShipyard,object.iNavalBase}}
unitTypeBuild[object.uSubmarineII.id] = {allImprovements={object.iShipyard,object.iNavalBase,object.iMilitaryFactory}}


--GENERIC LAND UNITS
unitTypeBuild[object.uWorkers.id] = {allImprovements={object.iLocalIndustry}}
--An empire's capital can train Engineers, even without a university or the required tech
unitTypeBuild[object.uEngineers.id] = {allImprovements=object.iOfficerAcademy,alternateParameters={engineerAltCondition}}
unitTypeBuild[object.uPartisans.id] = {allImprovements={object.iBarracks,object.iPropaganda}}
unitTypeBuild[object.uArtillery.id] = {allImprovements={object.iBarracks,object.iMilitaryFactory}}
unitTypeBuild[object.uHeavyArtillery.id] = {allImprovements={object.iBarracks,object.iMilitaryFactory}}
unitTypeBuild[object.uMobileArtillery.id] = {allImprovements={object.iBarracks,object.iMilitaryFactory}}
unitTypeBuild[object.uAABattery.id] = {allImprovements={object.iBarracks,object.iMilitaryFactory}}
unitTypeBuild[object.uAABatteryII.id] = {allImprovements={object.iBarracks,object.iAssemblyLine}}
unitTypeBuild[object.uTradeTruck.id] = {allImprovements={object.iAssemblyLine}}
unitTypeBuild[object.uArmoredInfantry.id] = {allImprovements={object.iBarracks,object.iAssemblyLine}}
unitTypeBuild[object.uLightTank.id] = {allImprovements={object.iBarracks,object.iMilitaryFactory}}
unitTypeBuild[object.uMediumTank.id] = {allImprovements={object.iBarracks,object.iAssemblyLine}}


--GENERIC AIR UNITS
unitTypeBuild[object.uFighter.id] = {allImprovements={object.iAirbase,object.iMilitaryFactory}}
unitTypeBuild[object.uJetFighter.id] = {allImprovements={object.iAirbase,object.iAssemblyLine}}
unitTypeBuild[object.uBomber.id] = {allImprovements={object.iAirbase,object.iMilitaryFactory}}
unitTypeBuild[object.uAlliedHeavyBomber.id] = {allImprovements={object.iAirbase,object.iAssemblyLine}}
unitTypeBuild[object.uAxisHeavyBomber.id] = {allImprovements={object.iAirbase,object.iAssemblyLine}}
unitTypeBuild[object.uAtomBomb.id] = {allImprovements={object.iResearchFacility,object.iArmamentsPlant}}


--RUSSIAN UNITS
--Soviet home cities with a Barracks can train Red Army Units
unitTypeBuild[object.uRedArmy.id] = {location=SovietHomeCities,allImprovements={object.iBarracks}}
unitTypeBuild[object.uRedArmyII.id] = {location=SovietHomeCities,allImprovements={object.iBarracks}}
unitTypeBuild[object.uShockTroops.id] = {location=SovietHomeCities,allImprovements={object.iBarracks}}

--Certain Soviet home cities with a Barracks and an Airbase can train Red Army Airborne Troops
unitTypeBuild[object.uAirborneTroops.id] = {location=SovAirborneCities,allImprovements={object.iBarracks,object.iAirbase}}

--Certain Soviet Hero cities with a Barracks can train Red Army Guards units
unitTypeBuild[object.uRedArmyGuards.id] = {location=GuardsHomeCities,allImprovements={object.iBarracks}}

--Certain Soviet Hero cities with a Barracks and a Tank Factory can train Red Army Guards tank units
unitTypeBuild[object.uGuardsT34.id] = {location=GuardsHomeCities,allImprovements={object.iBarracks,object.iTankFactory}}

--Soviet home cities with a Barracks and a Military Factory can train Red Army Katyusha units
unitTypeBuild[object.uKatyusha.id] = {location=SovietHomeCities,allImprovements={object.iBarracks,object.iMilitaryFactory}}

--Soviet home cities with a Barracks and a Military Factory can train the Red Army medium tank units
unitTypeBuild[object.uT34.id] = {location=SovietHomeCities,allImprovements={object.iBarracks,object.iMilitaryFactory}}
unitTypeBuild[object.uT3485.id] = {location=SovietHomeCities,allImprovements={object.iBarracks,object.iMilitaryFactory}}

--Soviet home cities with a Barracks and a Tank Factory can train the Red Army heavy tank units
unitTypeBuild[object.uKV1.id] = {location=SovietHomeCities,allImprovements={object.iBarracks,object.iTankFactory}}
unitTypeBuild[object.uIS2.id] = {location=SovietHomeCities,allImprovements={object.iBarracks,object.iTankFactory}}

--Soviet home cities with an Airbase and a Military Factory can train the early Red Air Force units
unitTypeBuild[object.uI16.id] = {location=SovietHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}
unitTypeBuild[object.uLa5.id] = {location=SovietHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}
unitTypeBuild[object.uTB3.id] = {location=SovietHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}

--Soviet home cities with an Airbase and an Assembly Line can train the later Red Air Force units
unitTypeBuild[object.uYak3.id] = {location=SovietHomeCities,allImprovements={object.iAirbase,object.iAssemblyLine}}
unitTypeBuild[object.uIl2Sturmovik.id] = {location=SovietHomeCities,allImprovements={object.iAirbase,object.iAssemblyLine}}


--JAPANESE UNITS
--Japanese occupied cities with a Barracks can train Imperial Armies
unitTypeBuild[object.uImperialArmy.id] = {location=JapaneseArmyCities,allImprovements={object.iBarracks}}

--Certain Japanese cities with a Barracks and an Airbase can train Imperial Paratroops
unitTypeBuild[object.uIJAParatroops.id] = {location=JapaneseParaCities,allImprovements={object.iBarracks,object.iAirbase}}

--Certain Japanese cities with Barracks and a Naval Base can train Imperial Marines
unitTypeBuild[object.uSNLFMarines.id] = {location=JapaneseMarinesCities,allImprovements={object.iBarracks,object.iNavalBase}}

--Japanese home cities with a Barracks can train Bushido soldiers
unitTypeBuild[object.uAssaultTroops.id] = {location=JapaneseHomeCities,allImprovements={object.iBarracks}}

--Japanese home cities with a Barracks and a Military Factory can train the Imperial Army tank units
unitTypeBuild[object.uType97ChiHa.id] = {location=JapaneseHomeCities,allImprovements={object.iBarracks,object.iMilitaryFactory}}
unitTypeBuild[object.uType3ChiNu.id] = {location=JapaneseHomeCities,allImprovements={object.iBarracks,object.iMilitaryFactory}}
unitTypeBuild[object.uType5ChiRi.id] = {location=JapaneseHomeCities,allImprovements={object.iBarracks,object.iMilitaryFactory}}

--Japanese home cities with an Airbase and a Military Factory can train the early Imperial Air units
unitTypeBuild[object.uA6MZero.id] = {location=JapaneseHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}
unitTypeBuild[object.uKi61Hien.id] = {location=JapaneseHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}
unitTypeBuild[object.uD3A.id] = {location=JapaneseHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}
unitTypeBuild[object.uKi67Hiryu.id] = {location=JapaneseHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}

--Japanese home cities with an Airbase and an Assembly Line can train the later Imperial Air units
unitTypeBuild[object.uKi84Hayate.id] = {location=JapaneseHomeCities,allImprovements={object.iAirbase,object.iAssemblyLine}}
unitTypeBuild[object.uD4YSuisei.id] = {location=JapaneseHomeCities,allImprovements={object.iAirbase,object.iAssemblyLine}}

--Japanese home cities with an Airbase can train Kamikaze units
unitTypeBuild[object.uMXY7Ohka.id] = {location=JapaneseHomeCities,allImprovements={object.iAirbase}}


--GERMAN UNITS
--German home cities with a Barracks can train Grenadier units
unitTypeBuild[object.uGrenadiers.id] = {location=GermanHomeCities,allImprovements={object.iBarracks}}
unitTypeBuild[object.uGrenadiersII.id] = {location=GermanHomeCities,allImprovements={object.iBarracks}}

--If Berlin has a Barracks and an Officer Academy, the city can train Waffen SS units
unitTypeBuild[object.uWaffenSS.id] = {location=WaffenSSCities,allImprovements={object.iBarracks,object.iOfficerAcademy}}

--If Berlin or Frankfurt have a Barracks and an Airbase, these cities can train Fallschirmjager units
unitTypeBuild[object.uFallschirmjagers.id] = {location=FallschirmHomeCities,allImprovements={object.iBarracks,object.iAirbase}}

--German home cities with a Barracks and a Military Factory can train the early Panzer units
unitTypeBuild[object.uPanzergrenadiers.id] = {location=GermanHomeCities,allImprovements={object.iBarracks,object.iMilitaryFactory}}
unitTypeBuild[object.uPzKpfwIII.id] = {location=GermanHomeCities,allImprovements={object.iBarracks,object.iMilitaryFactory}}
unitTypeBuild[object.uPzKpfwIV.id] = {location=GermanHomeCities,allImprovements={object.iBarracks,object.iMilitaryFactory}}

--German home cities with a Barracks and assembly lines can train the later Panzer units
unitTypeBuild[object.uPzKpfwVPanther.id] = {location=GermanHomeCities,allImprovements={object.iBarracks,object.iAssemblyLine}}
unitTypeBuild[object.uPzKpfwVITiger.id] = {location=GermanHomeCities,allImprovements={object.iBarracks,object.iAssemblyLine}}

--German home cities with an Airbase and a Military Factory can train the early Luftwaffe units
unitTypeBuild[object.uBf109.id] = {location=GermanHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}
unitTypeBuild[object.uJu87Stuka.id] = {location=GermanHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}
unitTypeBuild[object.uJu88.id] = {location=GermanHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}

--German home cities with an Airbase and an Assembly Line can train the later Luftwaffe units
unitTypeBuild[object.uFw190.id] = {location=GermanHomeCities,allImprovements={object.iAirbase,object.iAssemblyLine}}
unitTypeBuild[object.uHs129.id] = {location=GermanHomeCities,allImprovements={object.iAirbase,object.iAssemblyLine}}

--German home cities with an an Airbase, an Assembly Line and an Armament Plant can train the Luftwaffe Amerika bomber units
unitTypeBuild[object.uJu390.id] = {location=GermanHomeCities,allImprovements={object.iAirbase,object.iAssemblyLine,object.iArmamentsPlant}}

--German home cities with an Airbase and a Research Facility can train the advanced Luftwaffe units
unitTypeBuild[object.uMe262Schwalbe.id] = {location=GermanHomeCities,allImprovements={object.iAirbase,object.iResearchFacility}}
unitTypeBuild[object.uV2Missile.id] = {location=GermanHomeCities,allImprovements={object.iAirbase,object.iResearchFacility}}


--ITALIAN UNITS
--Control of Italian home cities with a Barracks will allow the training of Italian Armies
unitTypeBuild[object.uItalianArmy.id] = {location=ItalianHomeCities,allImprovements={object.iBarracks}}

--FINNISH UNITS
--Control of Finnish home cities with a Barracks will allow the training of Finnish Armies
unitTypeBuild[object.uFinnishArmy.id] = {location=FinnHomeCities,allImprovements={object.iBarracks}}

--ROMANIAN UNITS
--Control of Romanian home cities with a Barracks will allow the training of Romanian Armies
unitTypeBuild[object.uRomanianArmy.id] = {location=RomanianHomeCities,allImprovements={object.iBarracks}}

--FRENCH UNITS
--Germany can train Vichy Armies in certain French cities with a Barracks
unitTypeBuild[object.uVichyArmy.id] = {location=VichyHomeCities,allImprovements={object.iBarracks}}


--SPANISH UNITS
--Control of Spanish home cities with a Barracks will allow the training of Spanish Armies
unitTypeBuild[object.uSpanishArmy.id] = {location=SpanishHomeCities,allImprovements={object.iBarracks}}

--DUTCH UNITS
--Control of certain Dutch Far East cities with a Barracks will allow the training of Dutch Armies
unitTypeBuild[object.uDutchArmy.id] = {location=DutchHomeCities,allImprovements={object.iBarracks}}

--GARRISON UNITS
--Certain neutral cities with a Barracks will allow the training of Garrisons
unitTypeBuild[object.uGarrison.id] = {location=GarrisonHomeCities,allImprovements={object.iBarracks}}

--MUSLIM UNITS
--Certain Muslim cities with a Barracks will allow the training of Muslim Armies
unitTypeBuild[object.uMuslimArmy.id] = {location=MuslimHomeCities,allImprovements={object.iBarracks}}

--COLONIAL RIFLE UNITS
--Certain Eastern cities with a Barracks will allow the training of Colonial Rifles
unitTypeBuild[object.uColonialRifles.id] = {location=ColonialHomeCities,allImprovements={object.iBarracks}}


--AMERICAN UNITS
--American home cities with a Barracks can train US Army units
unitTypeBuild[object.uUSArmy.id] = {location=USAHomeCities,allImprovements={object.iBarracks}}
unitTypeBuild[object.uUSArmyII.id] = {location=USAHomeCities,allImprovements={object.iBarracks}}

--Certain American home cities with a Barracks and an Airbase can train US Airborne units
unitTypeBuild[object.uUSAirborne.id] = {location=USAirborneCities,allImprovements={object.iBarracks,object.iAirbase}}

--Certain American home cities with a Barracks and a Naval Base can train US Marine units
unitTypeBuild[object.uUSMarines.id] = {location=USMarineCities,allImprovements={object.iBarracks,object.iNavalBase}}

--Certain American home cities with a Naval Base and a Military Factory can train US Marine LVT units
unitTypeBuild[object.uUSMCLVT.id] = {location=USMarineCities,allImprovements={object.iNavalBase,object.iMilitaryFactory}}

--American home cities with a Barracks and a Military Factory can train the early US Army tanks
unitTypeBuild[object.uM3Lee.id] = {location=USAHomeCities,allImprovements={object.iBarracks,object.iMilitaryFactory}}

--American home cities with a Barracks and an Assembly Line can train the later US Army tanks
unitTypeBuild[object.uM4Sherman.id] = {location=USAHomeCities,allImprovements={object.iBarracks,object.iAssemblyLine}}
unitTypeBuild[object.uM26Pershing.id] = {location=USAHomeCities,allImprovements={object.iBarracks,object.iAssemblyLine}}

--American home cities with an Airbase and a Military Factory can train the early USAAF units
unitTypeBuild[object.uP38Lightning.id] = {location=USAHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}
unitTypeBuild[object.uB17Fortress.id] = {location=USAHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}

--American home cities with an Airbase and an Assembly Line can train the later USAAF units
unitTypeBuild[object.uP47Thunderbolt.id] = {location=USAHomeCities,allImprovements={object.iAirbase,object.iAssemblyLine}}
unitTypeBuild[object.uP51Mustang.id] = {location=USAHomeCities,allImprovements={object.iAirbase,object.iAssemblyLine}}

--American home cities with an Airbase, an Assembly Line and an Armament Plant can train the USAAF super bomber units
unitTypeBuild[object.uB29SuperFortress.id] = {location=USAHomeCities,allImprovements={object.iAirbase,object.iAssemblyLine,object.iArmamentsPlant}}

--American home cities with an Airbase and a Research Facility can train the advanced USAAF units
unitTypeBuild[object.uP80ShootingStar.id] = {location=USAHomeCities,allImprovements={object.iAirbase,object.iResearchFacility}}

--Certain American cities with an Airbase and a Naval Base can train F6F Hellcat units
unitTypeBuild[object.uF6FHellcat.id] = {location=USNAFBases,allImprovements={object.iAirbase,object.iNavalBase}}


--BRITISH UNITS
--British home cities with a Barracks can train British Infantry units
unitTypeBuild[object.uInfantry.id] = {location=BritishHomeCities,allImprovements={object.iBarracks}}
unitTypeBuild[object.uInfantryII.id] = {location=BritishHomeCities,allImprovements={object.iBarracks}}

--The cities of Newcastle and Birmingham with a Barracks and an Airbase can train British Paratrooper units
unitTypeBuild[object.uParatroopers.id] = {location=BritParaCities,allImprovements={object.iBarracks,object.iAirbase}}

--The cities of London and Glasgow with a Barracks and a Naval Base can train British Commando units
unitTypeBuild[object.uCommandos.id] = {location=BritCommandoCities,allImprovements={object.iBarracks,object.iNavalBase}}

--Certain British Indian cities with a Barracks can train Gurkha regiments
unitTypeBuild[object.uGurkhas.id] = {location=GurkhaHomeCities,allImprovements={object.iBarracks}}

--British Australian cities with a Barracks can train Anzac units
unitTypeBuild[object.uAnzacs.id] = {location=AnzacHomeCities,allImprovements={object.iBarracks}}

--British Canadian cities with a Barracks can train Canadian army units
unitTypeBuild[object.uCanadianArmy.id] = {location=CanadianHomeCities,allImprovements={object.iBarracks}}

--British home cities with a Barracks and a Military Factory can train the early British tanks
unitTypeBuild[object.uValentine.id] = {location=BritishHomeCities,allImprovements={object.iBarracks,object.iMilitaryFactory}}
unitTypeBuild[object.uChurchill.id] = {location=BritishHomeCities,allImprovements={object.iBarracks,object.iMilitaryFactory}}

--British home cities with a Barracks and an Assembly Line can train the later British tanks
unitTypeBuild[object.uComet.id] = {location=BritishHomeCities,allImprovements={object.iBarracks,object.iAssemblyLine}}
unitTypeBuild[object.uCenturion.id] = {location=BritishHomeCities,allImprovements={object.iBarracks,object.iAssemblyLine}}

--British home cities with an Airbase and a Military Factory can train the early RAF units
unitTypeBuild[object.uHurricane.id] = {location=BritishHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}
unitTypeBuild[object.uSpitfire.id] = {location=BritishHomeCities,allImprovements={object.iAirbase,object.iMilitaryFactory}}

--British home cities with an Airbase and an Assembly Line can train the later RAF units
unitTypeBuild[object.uTempest.id] = {location=BritishHomeCities,allImprovements={object.iAirbase,object.iAssemblyLine}}
unitTypeBuild[object.uMosquito.id] = {location=BritishHomeCities,allImprovements={object.iAirbase,object.iAssemblyLine}}
unitTypeBuild[object.uLancaster.id] = {location=BritishHomeCities,allImprovements={object.iAirbase,object.iAssemblyLine}}

--British home cities with an Airbase and a Research Facility can train the advanced RAF units
unitTypeBuild[object.uMeteor.id] = {location=BritishHomeCities,allImprovements={object.iAirbase,object.iResearchFacility}}


--CHINESE UNITS
--Chinese republican cities (including liberated ones) with a Barracks can train KMT army units
unitTypeBuild[object.uKMTArmy.id] = {location=ChineseHomeCities,allImprovements={object.iBarracks}}
unitTypeBuild[object.uKMTArmyII.id] = {location=ChineseHomeCities,allImprovements={object.iBarracks}}

--Two KMT cities with airbases can ship in allied lend-lease AVG fighters
unitTypeBuild[object.uAVGWarhawk.id] = {location=ChineseAVGBases,allImprovements={object.iAirbase}}

--Chinese cities with a Barracks can train PLA troops when captured by rebels or soviets
unitTypeBuild[object.uPLATroops.id] = {location=ChineseHomeCities,allImprovements={object.iBarracks}}

local improvementBuild = {}

--INDUSTRY IMPROVEMENTS
--Shipyards can be built in cities with Local Industry and Docks
improvementBuild[object.iShipyard.id] = {someImprovements={object.iLocalIndustry,object.iDocks},numberOfImprovements=2}

--Military Factories can be built in cities with Local Industry and a Rail Station
improvementBuild[object.iMilitaryFactory.id] = {someImprovements={object.iLocalIndustry,object.iRailStation},numberOfImprovements=2}

--Naval Bases can be built in cities with a Barracks and Docks
improvementBuild[object.iNavalBase.id] = {someImprovements={object.iBarracks,object.iDocks},numberOfImprovements=2}

--Airbases can be built in cities with a Barracks and Officer Academy
improvementBuild[object.iAirbase.id] = {someImprovements={object.iBarracks,object.iOfficerAcademy},numberOfImprovements=2}

--Financial Districts can only be built in cities at size 10 with Warehouses and a Freight Depot
improvementBuild[object.iFinancialDistrict.id] = {someImprovements={object.iWarehouses,object.iFreightDepot},numberOfImprovements=2,minimumPopulation=10}

--DEFENCE IMPROVEMENTS
--City Defences must be built in cities with existing Barracks
improvementBuild[object.iCityDefences.id] = {someImprovements={object.iBarracks}}

--AA Batteries must be built in cities with existing City Defences
improvementBuild[object.iAABattery.id] = {someImprovements={object.iCityDefences}}

--Coastal Batteries must be built in coastal cities with existing City Defences
improvementBuild[object.iCoastalBattery.id] = {someImprovements={object.iCityDefences}}

--Wasserfall Batteries must be built in cities with an existing AA Battery
improvementBuild[object.iWasserfallBattery.id] = {someImprovements={object.iAABattery}}

--CIVIC ORDER IMPROVEMENTS
--Military Police can be created in cities with Barracks and a Governor
improvementBuild[object.iMilitaryPolice.id] = {someImprovements={object.iBarracks,object.iGovernor},numberOfImprovements=2}

--Cinemas can be built in cities with existing Propaganda
improvementBuild[object.iCinema.id] = {someImprovements={object.iPropaganda}}

local wonderBuild = {}

--The National Shipyard Wonder can only be constructed in a city with a Shipyard and an Assembly Lines
wonderBuild[object.wNationalShipyard.id] = {allImprovements={object.iShipyard,object.iAssemblyLine}}


canBuildFunctions.supplyUnitTypeParameters(unitTypeBuild)
canBuildFunctions.supplyImprovementParameters(improvementBuild)
canBuildFunctions.supplyWonderParameters(wonderBuild)
